@@ -10,23 +10,23 @@ use std::sync::Arc;
 use async_trait::async_trait;
 use collate::Collate;
 use destream::{de, en};
-use futures::future::{self, try_join, try_join_all, Future, TryFutureExt};
+use futures::future::{self, Future, TryFutureExt};
 use futures::join;
-use futures::stream::{self, FuturesOrdered, TryStreamExt};
+use futures::stream::{self, FuturesOrdered, FuturesUnordered, TryStreamExt};
 use log::debug;
 use uuid::Uuid;
 
 use tc_error::*;
 use tc_transact::fs::*;
-use tc_transact::lock::{Mutable, TxnLock};
+use tc_transact::lock::TxnLock;
 use tc_transact::{Transact, Transaction, TxnId};
 use tc_value::{Value, ValueCollator};
-use tcgeneric::{Instance, TCBoxTryFuture, TCTryStream, Tuple};
+use tcgeneric::{Instance, TCBoxTryFuture, TCBoxTryStream, Tuple};
 
-use super::{BTree, BTreeInstance, BTreeSlice, BTreeType, Key, Range, RowSchema};
+use super::{BTree, BTreeInstance, BTreeSlice, BTreeType, BTreeWrite, Key, Range, RowSchema};
 
 type Selection<'a> = FuturesOrdered<
-    Pin<Box<dyn Future<Output = TCResult<TCTryStream<'a, Key>>> + Send + Unpin + 'a>>,
+    Pin<Box<dyn Future<Output = TCResult<TCBoxTryStream<'a, Key>>> + Send + Unpin + 'a>>,
 >;
 
 const DEFAULT_BLOCK_SIZE: usize = 4_000;
@@ -215,7 +215,7 @@ struct Inner<F, D, T> {
     schema: RowSchema,
     order: usize,
     collator: ValueCollator,
-    root: TxnLock<Mutable<NodeId>>,
+    root: TxnLock<NodeId>,
     dir: PhantomData<D>,
     txn: PhantomData<T>,
 }
@@ -274,7 +274,7 @@ where
             let collator = &self.inner.collator;
             let file = &self.inner.file;
 
-            let node = file.read_block(txn_id, node_id).await?;
+            let mut node = file.write_block(txn_id, node_id).await?;
 
             #[cfg(debug_assertions)]
             node.validate(&self.inner.schema, range);
@@ -285,11 +285,6 @@ where
             debug!("delete from {} [{}..{}] ({:?})", *node, l, r, range);
 
             if node.leaf {
-                if l == r {
-                    return Ok(());
-                }
-
-                let mut node = node.upgrade(file).await?;
                 for i in l..r {
                     node.keys[i].deleted = true;
                 }
@@ -297,20 +292,22 @@ where
 
                 Ok(())
             } else if r > l {
-                let mut node = node.upgrade(file).await?;
-                let mut deletes = Vec::with_capacity(r - l);
+                node.rebalance = true;
 
                 for i in l..r {
                     node.keys[i].deleted = true;
-                    deletes.push(self._delete_range(txn_id, node.children[i].clone(), range));
                 }
-                node.rebalance = true;
 
-                let child_id = node.children[r].clone();
-                let last_delete = self._delete_range(txn_id, child_id, range);
-                try_join(try_join_all(deletes), last_delete).await?;
+                let deletes: FuturesUnordered<_> = node.children[l..(r + 1)]
+                    .iter()
+                    .cloned()
+                    .map(|child_id| self._delete_range(txn_id, child_id, range))
+                    .collect();
 
-                Ok(())
+                // free the write lock on this node before deleting from the child nodes
+                std::mem::drop(node);
+
+                deletes.try_fold((), |(), ()| future::ready(Ok(()))).await
             } else {
                 let child_id = node.children[r].clone();
                 self._delete_range(txn_id, child_id, range).await
@@ -318,11 +315,6 @@ where
         })
     }
 
-    pub(super) async fn delete_range(&self, txn_id: TxnId, range: &Range) -> TCResult<()> {
-        let root_id = self.inner.root.read(&txn_id).await?;
-        self._delete_range(txn_id, (*root_id).clone(), range).await
-    }
-
     fn _insert(
         &self,
         txn_id: TxnId,
@@ -335,22 +327,29 @@ where
             let order = self.inner.order;
 
             let i = collator.bisect_left(&node.keys, &key);
-            if i < node.keys.len() && collator.compare_slice(&node.keys[i], &key) == Ordering::Equal
-            {
-                if node.keys[i].deleted {
-                    let mut node = node.upgrade(file).await?;
-                    node.keys[i].deleted = false;
-                }
-
-                return Ok(());
-            }
 
             #[cfg(debug_assertions)]
             debug!("insert at index {} into {}", i, *node);
 
             if node.leaf {
+                let key = NodeKey::new(key);
                 let mut node = node.upgrade(file).await?;
-                node.keys.insert(i, NodeKey::new(key));
+
+                if i == node.keys.len() {
+                    node.keys.insert(i, key);
+                    return Ok(());
+                }
+
+                match collator.compare_slice(&key, &node.keys[i]) {
+                    Ordering::Less => node.keys.insert(i, key),
+                    Ordering::Equal => {
+                        #[cfg(debug_assertions)]
+                        debug!("un-delete key at {}: {}", i, key);
+                        node.keys[i].deleted = false
+                    }
+                    Ordering::Greater => panic!("error in Collate::bisect_left"),
+                }
+
                 Ok(())
             } else {
                 let child_id = node.children[i].clone();
@@ -393,7 +392,7 @@ where
         txn_id: TxnId,
         node: B,
         range: Range,
-    ) -> TCResult<TCTryStream<'a, Key>>
+    ) -> TCResult<TCBoxTryStream<'a, Key>>
     where
         Self: 'a,
     {
@@ -426,7 +425,7 @@ where
 
                 if !node.keys[i].deleted {
                     let key_at_i = TCResult::Ok(node.keys[i].value.to_vec());
-                    let key_at_i: TCTryStream<Key> =
+                    let key_at_i: TCBoxTryStream<Key> =
                         Box::pin(stream::once(future::ready(key_at_i)));
 
                     selected.push(Box::pin(future::ready(Ok(key_at_i))));
@@ -450,7 +449,7 @@ where
         txn_id: TxnId,
         node: B,
         range: Range,
-    ) -> TCResult<TCTryStream<'a, Key>>
+    ) -> TCResult<TCBoxTryStream<'a, Key>>
     where
         Self: 'a,
     {
@@ -493,7 +492,7 @@ where
 
                 if !node.keys[i].deleted {
                     let key_at_i = TCResult::Ok(node.keys[i].value.to_vec());
-                    let key_at_i: TCTryStream<Key> =
+                    let key_at_i: TCBoxTryStream<Key> =
                         Box::pin(stream::once(future::ready(key_at_i)));
 
                     selected.push(Box::pin(future::ready(Ok(key_at_i))));
@@ -511,7 +510,7 @@ where
         txn_id: TxnId,
         range: Range,
         reverse: bool,
-    ) -> TCResult<TCTryStream<'a, Key>>
+    ) -> TCResult<TCBoxTryStream<'a, Key>>
     where
         Self: 'a,
     {
@@ -609,26 +608,57 @@ where
             .file
             .read_block(txn_id, (*root_id).clone())
             .await?;
+
         Ok(root.keys.is_empty())
     }
 
-    async fn delete(&self, txn_id: TxnId) -> TCResult<()> {
-        let mut root = self.inner.root.write(txn_id).await?;
+    async fn keys<'a>(self, txn_id: TxnId) -> TCResult<TCBoxTryStream<'a, Key>> {
+        self.rows_in_range(txn_id, Range::default(), false).await
+    }
 
-        self.inner.file.truncate(txn_id).await?;
+    fn validate_key(&self, key: Key) -> TCResult<Key> {
+        if key.len() != self.inner.schema.len() {
+            return Err(TCError::bad_request("invalid key length", Tuple::from(key)));
+        }
 
-        *root = self.inner.file.unique_id(&txn_id).await?;
+        key.into_iter()
+            .zip(&self.inner.schema)
+            .map(|(val, col)| {
+                val.into_type(col.dtype)
+                    .ok_or_else(|| TCError::bad_request("invalid value for column", &col.name))
+            })
+            .collect()
+    }
+}
 
-        self.inner
-            .file
-            .create_block(txn_id, (*root).clone(), Node::new(true, None))
-            .await?;
+#[async_trait]
+impl<F: File<Node>, D: Dir, T: Transaction<D>> BTreeWrite for BTreeFile<F, D, T>
+where
+    Self: Clone,
+    BTreeSlice<F, D, T>: 'static,
+{
+    async fn delete(&self, txn_id: TxnId, range: Range) -> TCResult<()> {
+        if range == Range::default() {
+            let mut root = self.inner.root.write(txn_id).await?;
 
-        Ok(())
+            self.inner.file.truncate(txn_id).await?;
+
+            *root = self.inner.file.unique_id(&txn_id).await?;
+
+            self.inner
+                .file
+                .create_block(txn_id, (*root).clone(), Node::new(true, None))
+                .await?;
+
+            return Ok(());
+        }
+
+        let root_id = self.inner.root.read(&txn_id).await?;
+        self._delete_range(txn_id, (*root_id).clone(), &range).await
     }
 
     async fn insert(&self, txn_id: TxnId, key: Key) -> TCResult<()> {
-        let key = validate_key(key, &self.inner.schema)?;
+        let key = self.validate_key(key)?;
 
         let file = &self.inner.file;
         let order = self.inner.order;
@@ -653,11 +683,7 @@ where
         #[cfg(debug_assertions)]
         debug!("root node {} is {}", *root_id, *root);
 
-        if root.leaf {
-            assert!(root.children.is_empty());
-        } else {
-            assert!(!root.children.is_empty());
-        }
+        assert_eq!(root.children.is_empty(), root.leaf);
 
         if root.keys.len() == (2 * order) - 1 {
             // split_child will need a write lock on root, so release the read lock
@@ -685,10 +711,6 @@ where
             self._insert(txn_id, root, key).await
         }
     }
-
-    async fn keys<'a>(self, txn_id: TxnId) -> TCResult<TCTryStream<'a, Key>> {
-        self.rows_in_range(txn_id, Range::default(), false).await
-    }
 }
 
 #[async_trait]
@@ -757,7 +779,7 @@ impl<F: File<Node>, D: Dir, T: Transaction<D>> Restore<D> for BTreeFile<F, D, T>
 impl<F: File<Node>, D: Dir, T: Transaction<D>, I: BTreeInstance + 'static> CopyFrom<D, I>
     for BTreeFile<F, D, T>
 {
-    async fn copy_from(source: I, file: F, txn: T) -> TCResult<Self> {
+    async fn copy_from(source: I, file: F, txn: &T) -> TCResult<Self> {
         let txn_id = *txn.id();
         let schema = source.schema().clone();
         let dest = Self::create(file, schema, txn_id).await?;
@@ -811,18 +833,3 @@ fn validate_schema(schema: &RowSchema) -> TCResult<usize> {
 
     Ok(order)
 }
-
-#[inline]
-fn validate_key(key: Key, schema: &RowSchema) -> TCResult<Key> {
-    if key.len() != schema.len() {
-        return Err(TCError::bad_request("invalid key length", Tuple::from(key)));
-    }
-
-    key.into_iter()
-        .zip(schema)
-        .map(|(val, col)| {
-            val.into_type(col.dtype)
-                .ok_or_else(|| TCError::bad_request("invalid value for column", &col.name))
-        })
-        .collect()
-}
