@@ -13,7 +13,7 @@ use safecast::*;
 use tc_error::*;
 use tc_transact::fs::{Dir, File, Hash};
 use tc_transact::{IntoView, Transaction, TxnId};
-use tc_value::{Link, NumberType, Value, ValueCollator, ValueType};
+use tc_value::{NumberType, Value, ValueCollator, ValueType};
 use tcgeneric::*;
 
 pub use file::{BTreeFile, Node};
@@ -22,6 +22,7 @@ pub use slice::BTreeSlice;
 mod file;
 mod slice;
 
+const ERR_VIEW_WRITE: &str = "BTree view does not support write operations";
 const PREFIX: PathLabel = path_label(&["state", "collection", "btree"]);
 
 /// The file extension of a [`BTree`] as stored on-disk
@@ -38,13 +39,13 @@ pub type Range = collate::Range<Value, Key>;
 pub trait BTreeInstance: Clone + Instance {
     type Slice: BTreeInstance;
 
-    /// Return a reference to this `BTree`'s collator.
+    /// Borrow this `BTree`'s collator.
     fn collator(&self) -> &ValueCollator;
 
-    /// Return a reference to this `BTree`'s schema.
+    /// Borrow to this `BTree`'s schema.
     fn schema(&self) -> &RowSchema;
 
-    /// Return a slice of this `BTree`'s with the given range.
+    /// Return a slice of this `BTree` with the given range.
     fn slice(self, range: Range, reverse: bool) -> TCResult<Self::Slice>;
 
     /// Return the number of [`Key`]s in this `BTree`.
@@ -58,19 +59,28 @@ pub trait BTreeInstance: Clone + Instance {
     /// Return `true` if this `BTree` has no [`Key`]s.
     async fn is_empty(&self, txn_id: TxnId) -> TCResult<bool>;
 
+    /// Return a `Stream` of this `BTree`'s [`Key`]s.
+    async fn keys<'a>(self, txn_id: TxnId) -> TCResult<TCBoxTryStream<'a, Key>>
+    where
+        Self: 'a;
+
+    /// Return an error if the given key does not match this `BTree`'s schema
+    ///
+    /// If the key is valid, this will return a copy with the data types correctly casted.
+    fn validate_key(&self, key: Key) -> TCResult<Key>;
+}
+
+/// [`BTree`] write methods.
+#[async_trait]
+pub trait BTreeWrite: BTreeInstance {
     /// Delete all the [`Key`]s in this `BTree`.
-    async fn delete(&self, txn_id: TxnId) -> TCResult<()>;
+    async fn delete(&self, txn_id: TxnId, range: Range) -> TCResult<()>;
 
     /// Insert the given [`Key`] into this `BTree`.
     ///
     /// If the [`Key`] is already present, this is a no-op.
     async fn insert(&self, txn_id: TxnId, key: Key) -> TCResult<()>;
 
-    /// Return a `Stream` of this `BTree`'s [`Key`]s.
-    async fn keys<'a>(self, txn_id: TxnId) -> TCResult<TCTryStream<'a, Key>>
-    where
-        Self: 'a;
-
     /// Insert all the keys from the given `Stream` into this `BTree`.
     ///
     /// This will stop and return an error if it encounters an invalid [`Key`].
@@ -91,7 +101,7 @@ impl<'en, F: File<Node>, D: Dir, T: Transaction<D>> Hash<'en, D> for BTree<F, D,
     type Item = Key;
     type Txn = T;
 
-    async fn hashable(&'en self, txn: &'en T) -> TCResult<TCTryStream<'en, Self::Item>> {
+    async fn hashable(&'en self, txn: &'en T) -> TCResult<TCBoxTryStream<'en, Self::Item>> {
         self.clone().keys(*txn.id()).await
     }
 }
@@ -250,12 +260,10 @@ impl de::FromStream for Column {
 
 impl<'en> en::IntoStream<'en> for Column {
     fn into_stream<E: en::Encoder<'en>>(self, encoder: E) -> Result<E::Ok, E::Error> {
-        let dtype = Value::Link(Link::from(self.dtype.path()));
-
         if let Some(max_len) = self.max_len {
-            (self.name, dtype.to_string(), max_len).into_stream(encoder)
+            (self.name, self.dtype, max_len).into_stream(encoder)
         } else {
-            (self.name, dtype.to_string()).into_stream(encoder)
+            (self.name, self.dtype).into_stream(encoder)
         }
     }
 }
@@ -376,21 +384,7 @@ where
         }
     }
 
-    async fn delete(&self, txn_id: TxnId) -> TCResult<()> {
-        match self {
-            Self::File(file) => file.delete(txn_id).await,
-            Self::Slice(slice) => slice.delete(txn_id).await,
-        }
-    }
-
-    async fn insert(&self, txn_id: TxnId, key: Key) -> TCResult<()> {
-        match self {
-            Self::File(file) => file.insert(txn_id, key).await,
-            Self::Slice(slice) => slice.insert(txn_id, key).await,
-        }
-    }
-
-    async fn keys<'a>(self, txn_id: TxnId) -> TCResult<TCTryStream<'a, Key>>
+    async fn keys<'a>(self, txn_id: TxnId) -> TCResult<TCBoxTryStream<'a, Key>>
     where
         Self: 'a,
     {
@@ -399,6 +393,33 @@ where
             Self::Slice(slice) => slice.keys(txn_id).await,
         }
     }
+
+    fn validate_key(&self, key: Key) -> TCResult<Key> {
+        match self {
+            Self::File(file) => file.validate_key(key),
+            Self::Slice(slice) => slice.validate_key(key),
+        }
+    }
+}
+
+#[async_trait]
+impl<F: File<Node>, D: Dir, T: Transaction<D>> BTreeWrite for BTree<F, D, T>
+where
+    Self: 'static,
+{
+    async fn delete(&self, txn_id: TxnId, range: Range) -> TCResult<()> {
+        match self {
+            Self::File(file) => file.delete(txn_id, range).await,
+            _ => Err(TCError::unsupported(ERR_VIEW_WRITE)),
+        }
+    }
+
+    async fn insert(&self, txn_id: TxnId, key: Key) -> TCResult<()> {
+        match self {
+            Self::File(file) => file.insert(txn_id, key).await,
+            _ => Err(TCError::unsupported(ERR_VIEW_WRITE)),
+        }
+    }
 }
 
 impl<F, D, T> From<BTreeFile<F, D, T>> for BTree<F, D, T> {
@@ -549,7 +570,7 @@ where
 /// A view of a [`BTree`] within a single [`Transaction`], used in serialization.
 pub struct BTreeView<'en> {
     schema: Vec<Column>,
-    keys: TCTryStream<'en, Key>,
+    keys: TCBoxTryStream<'en, Key>,
 }
 
 impl<'en> en::IntoStream<'en> for BTreeView<'en> {
